# encoding: utf-8

from __future__ import print_function, unicode_literals

import re
from collections import namedtuple


class CqlSplitter(object):
    """
    Makeshift CQL parser that can only split up multiple statements.

    C* does not accept multiple DDL queries as a single string, as it can with
    DML queries using batches. Hence, we must split up CQL files to run each
    statement individually. Do that by using a simple Regex scanner, that just
    recognizes strings, comments and delimiters, which is enough to split up
    statements without tripping when semicolons are commented or escaped.
    """

    Token = namedtuple('Token', 'tpe token')

    LINE_COMMENT = 1
    BLOCK_COMMENT = 2
    STRING = 3
    SEMICOLON = 4
    OTHER = 5
    WHITESPACE = 6

    @classmethod
    def scanner(cls):
        if not getattr(cls, '_scanner', None):
            def h(tpe):
                return lambda sc, tk: cls.Token(tpe, tk)

            cls._scanner = re.Scanner([
                (r"(--|//).*?$",               h(cls.LINE_COMMENT)),
                (r"\/\*.+?\*\/",               h(cls.BLOCK_COMMENT)),
                (r'"(?:[^"\\]|\\.)*"',         h(cls.STRING)),
                (r"'(?:[^'\\]|\\.)*'",         h(cls.STRING)),
                (r"\$\$(?:[^\$\\]|\\.)*\$\$",  h(cls.STRING)),
                (r";",                         h(cls.SEMICOLON)),
                (r"\s+",                       h(cls.WHITESPACE)),
                (r".",                         h(cls.OTHER))
            ], re.MULTILINE | re.DOTALL)
        return cls._scanner

    @classmethod
    def split(cls, query):
        """Split up content, and return individual statements uncommented"""
        tokens, match = cls.scanner().scan(query)
        cur_statement = ''
        statements = []

        for i, tk in enumerate(tokens):
            if tk.tpe == cls.LINE_COMMENT:
                pass
            elif tk.tpe == cls.SEMICOLON:
                stm = cur_statement.strip()
                if stm:
                    statements.append(stm)
                cur_statement = ''
            elif tk.tpe in (cls.WHITESPACE, cls.BLOCK_COMMENT):
                cur_statement += ' '
            elif tk.tpe in (cls.STRING, cls.OTHER):
                cur_statement += tk.token

        stm = cur_statement.strip()
        if stm:
            statements.append(stm)

        return statements



#!/bin/env python
# encoding: utf-8

from __future__ import unicode_literals

from yaml import Loader, SafeLoader
from cstarmigrate.cli import main


# Force YAML to load strings as unicode
def construct_yaml_str(self, node):
    # Override the default string handling function
    # to always return unicode objects
    return self.construct_scalar(node)


Loader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)
SafeLoader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)


if __name__ == '__main__':
    main()



# encoding: utf-8

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import sys
import os
import logging
import argparse
import subprocess

from cstarmigrate import (Migrator, Migration, MigrationConfig,
                          MigrationError)


def open_file(filename):
    if sys.platform == 'win32':
        os.startfile(filename)
    else:
        if 'XDG_CURRENT_DESKTOP' in os.environ:
            opener = ['xdg-open']
        elif 'EDITOR' in os.environ:
            opener = [os.environ['EDITOR']]
        else:
            opener = ['vi']

        opener.append(filename)
        subprocess.call(opener)


def main():
    logging.basicConfig(level=logging.INFO)
    logging.getLogger("cassandra.policies").setLevel(logging.ERROR)

    parser = argparse.ArgumentParser(
        description='Simple Cassandra migration tool')
    parser.add_argument('-H', '--hosts', default='127.0.0.1',
                        help='Comma-separated list of contact points')
    parser.add_argument('-p', '--port', type=int, default=9042,
                        help='Connection port')
    parser.add_argument('-u', '--user',
                        help='Connection username')
    parser.add_argument('-P', '--password',
                        help='Connection password')
    parser.add_argument('-c', '--config-file', default='cstar-migrate.yml',
                        help='Path to configuration file')
    parser.add_argument('-m', '--profile', default='dev',
                        help='Name of keyspace profile to use')
    parser.add_argument('-s', '--ssl-cert', default=None,
                        help="""
                        File path of .pem or .crt containing certificate of the
                        cassandra host you are connecting to (or the
                        certificate of the CA that signed the host certificate).
                        If this option is provided, cstar-migrate will use
                        ssl to connect to the cluster. If this option is not
                        provided, the -k and -t options will be ignored. """)
    parser.add_argument('-k', '--ssl-client-private-key', default=None,
                        help="""
                        File path of the .key file containing the private key
                        of the host on which the cstar-migrate command is
                        run. This option must be used in conjuction with the
                        -t option. This option is ignored unless the -s
                        option is provided.""")
    parser.add_argument('-t', '--ssl-client-cert', default=None,
                        help="""
                        File path of the .crt file containing the public
                        certificate of the host on which the cstar-migrate
                        command is run. This certificate (or the CA that signed
                        it) must be trusted by the cassandra host that
                        migrations are run against. This option must be used in
                        conjuction with the -k option. This option is ignored
                        unless the -s option is provided.""")
    parser.add_argument('-y', '--assume-yes', action='store_true',
                        help='Automatically answer "yes" for all questions')

    cmds = parser.add_subparsers(help='sub-command help')

    bline = cmds.add_parser(
        'baseline',
        help='Baseline database state, advancing migration information without '
             'making changes')
    bline.set_defaults(action='baseline')

    reset = cmds.add_parser(
        'reset',
        help='Reset database state, by dropping the keyspace (if it exists) '
             'and recreating it from scratch')
    reset.set_defaults(action='reset')

    mgrat = cmds.add_parser(
        'migrate',
        help='Migrate database up to the most recent (or specified) version '
             'by applying any new migration scripts in sequence')
    mgrat.add_argument('-f', '--force', action='store_true',
                       help='Force migration even if last attempt failed')
    mgrat.set_defaults(action='migrate')

    stats = cmds.add_parser(
        'status',
        help='Print current state of keyspace')
    stats.set_defaults(action='status')

    genrt = cmds.add_parser(
        'generate',
        help='Generate a new migration file')
    genrt.add_argument(
        'description',
        help='Brief description of the new migration')
    genrt.add_argument(
        '--python',
        dest='migration_type',
        action='store_const',
        const='python',
        default='cql',
        help='Generates a Python script instead of CQL.')

    genrt.set_defaults(action='generate')

    for sub in (bline, reset, mgrat):
        sub.add_argument('db_version', metavar='VERSION', nargs='?',
                         help='Database version to baseline/reset/migrate to')

    opts = parser.parse_args()
    # enable user confirmation if we're running the script from a TTY
    opts.cli_mode = sys.stdin.isatty()
    config = MigrationConfig.load(opts.config_file)

    if opts.action == 'generate':
        new_path = Migration.generate(config=config,
                                      description=opts.description,
                                      output=opts.migration_type)
        if sys.stdin.isatty():
            open_file(new_path)

        print(os.path.basename(new_path))
    else:
        with Migrator(config=config, profile=opts.profile,
                      hosts=opts.hosts.split(','), port=opts.port,
                      user=opts.user, password=opts.password,
                      host_cert_path=opts.ssl_cert,
                      client_key_path=opts.ssl_client_private_key,
                      client_cert_path=opts.ssl_client_cert) as migrator:
            cmd_method = getattr(migrator, opts.action)
            if not callable(cmd_method):
                print('Error: invalid command', file=sys.stderr)
                sys.exit(1)

            try:
                cmd_method(opts)
            except MigrationError as e:
                print('Error: {}'.format(e), file=sys.stderr)
                sys.exit(1)



# encoding: utf-8
# flake8: noqa: E402,F401

from __future__ import print_function, unicode_literals


class MigrationError(RuntimeError):
    """Base class for migration errors"""
    pass


class FailedMigration(MigrationError):
    """Database state contains failed migrations"""

    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(FailedMigration, self).__init__(
            'Migration failed, cannot continue '
            '(version {}): {}'.format(version, name))


class ConcurrentMigration(MigrationError):
    """Database state contains failed migrations"""

    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(ConcurrentMigration, self).__init__(
            'Migration already in progress '
            '(version {}): {}'.format(version, name))


class InconsistentState(MigrationError):
    """Database state differs from specified migrations"""

    def __init__(self, migration, version):
        self.migration = migration
        self.version = version

        super(InconsistentState, self).__init__(
            'Found inconsistency between specified migration and stored '
            'version: {} != {}'.format(migration, version))


class UnknownMigration(MigrationError):
    """Database contains migrations that have not been specified"""
    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(UnknownMigration, self).__init__(
            'Found version in database without corresponding '
            'migration (version {}): {} '.format(version, name))



from .migration import Migration
from .config import MigrationConfig
from .migrator import Migrator
